/*
 * Uniform Grid and Pair Management
 * Author: Peter Kohut
 *
 * press 'a' or '+' to add random AABB
 * press 'r' or '-' to remove AABB with highest ID(last)
 * press SPACE or ENTER to pause
 * press Esc or 'q' to exit application
 */

#include <windows.h>
#include <GL/glut.h>
#include <stdlib.h>
#include <time.h>

struct AABB {   // structure for singe AABB
    int ID, x, y, s, xvel, yvel;
};

struct cell {   // structure for cell in uniform grid
    int count;
    bool IDs [50];
};

struct poz {    // structure used in get_cell method
    int x, y;
};

int max = 10;   // current number of AABBs
AABB n [50];   // array of AABBs
cell grid [20][16];     // uniform grid 20x16, one cell 30px
int tick = 0;   // value for timer
bool pause = false;

/* GLUT callback Handlers */

static void resize(int width, int height)
{
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0f,width,height,0.0f,-1.0f,1.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity() ;
}

poz get_cell(int x, int y)
{   // get cell for corresponding point [x,y]
    poz temp;
    temp.x = x/30;
    temp.y = y/30;
    if(temp.x>19) temp.x = 19;
    if(temp.y>15) temp.y = 15;
    return temp;
}



void AddBox(int ID)
{   // add AABB
    poz min, max;
    min = get_cell(n[ID].x-n[ID].s,n[ID].y-n[ID].s);
    max = get_cell(n[ID].x+n[ID].s,n[ID].y+n[ID].s);
    for(int i=min.x;i<=max.x;i++){
        for(int j=min.y;j<=max.y;j++){
            grid[i][j].count++;
            grid[i][j].IDs[ID] = true;
        }
    }
}

void RemoveBox(int ID)
{   // remove AABB
    poz min, max;
    min = get_cell(n[ID].x-n[ID].s,n[ID].y-n[ID].s);
    max = get_cell(n[ID].x+n[ID].s,n[ID].y+n[ID].s);
    for(int i=min.x;i<=max.x;i++){
        for(int j=min.y;j<=max.y;j++){
            grid[i][j].count--;
            grid[i][j].IDs[ID] = false;
        }
    }
}

void UpdateBox(int ID)
{   // updates AABBs
    RemoveBox(ID);
    n[ID].x+= n[ID].xvel;   // apply x velocity
    n[ID].y+= n[ID].yvel;   // apply y velocity
    AddBox(ID);
}

void AddAABB()
{   // add new AABB with random values
    if(max<50){
        n[max].ID = max;
        n[max].x = rand() % 520 + 40;
        n[max].y = rand() % 400 + 40;
        n[max].s = rand() % 20 + 20;
        n[max].xvel = (rand() % 5 + 1) - (rand() % 5 + 1);
        n[max].yvel = (rand() % 5 + 1) - (rand() % 5 + 1);
        AddBox(max);
        max++;
    }
}

void RemoveAABB()
{   // remove AABB with highest ID
    if(max>0){
        max--;
        RemoveBox(max);
    }
}

void init()
{   // initialize AABBs
    for(int i=0;i<max;i++){
        n[i].ID = i;
        n[i].x = rand() % 520 + 40;
        n[i].y = rand() % 400 + 40;
        n[i].s = rand() % 20 + 20;
        n[i].xvel = (rand() % 5 + 1) - (rand() % 5 + 1);
        n[i].yvel = (rand() % 5 + 1) - (rand() % 5 + 1);
        AddBox(i);
    }
}

void bound_check()
{   // keeps AABBs in boundary by negating their velocity
    for(int i=0;i<max;i++){
        if(n[i].x-n[i].s<0) n[i].xvel = -n[i].xvel;
        if(n[i].x+n[i].s>600) n[i].xvel = -n[i].xvel;
        if(n[i].y-n[i].s<0) n[i].yvel = -n[i].yvel;
        if(n[i].y+n[i].s>480) n[i].yvel = -n[i].yvel;
    }
}

static void display(void)
{
    const double t = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
    int temp = t*90.0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if(tick<temp && !pause){
        for(int i=0;i<max;i++){
            UpdateBox(i);
        }
        bound_check();
        tick = temp;
    }

    // UNIFORM GRID
    glLoadIdentity();
    glTranslatef(0,0,-0.1);
    glColor4d(0,0,0,0.5);
    for(int i=1;i<20;i++){
        glBegin(GL_LINES);
            glVertex2d(30*i,0);
            glVertex2d(30*i,480);
        glEnd();
        if(i<16){
            glBegin(GL_LINES);
                glVertex2d(0,30*i);
                glVertex2d(600,30*i);
            glEnd();
        }
    }

    // highlight potential collosions
    for(int i=0;i<20;i++){
        for(int j=0;j<16;j++){
            if(grid[i][j].count<1){
                continue;
            }else{
                glLoadIdentity();
                if(grid[i][j].count>1) glColor4d(0.8,0.8,0,0.3);
                else glColor4d(0,0,0.8,0.3);
                glTranslatef(i*30+15,j*30+14,-0.01);
                glBegin(GL_QUADS);
                    glVertex2d(-14,-14);
                    glVertex2d(15,-14);
                    glVertex2d(15,15);
                    glVertex2d(-14,15);
                glEnd();
            }
        }
    }

    // AABBs
    glColor4d(0.8,0,0,0.65);
    for(int i=0;i<max;i++){
        glLoadIdentity();
        glTranslatef(n[i].x,n[i].y,0.01*i);
        glBegin(GL_QUADS);
            glVertex2d(-n[i].s,-n[i].s);
            glVertex2d(n[i].s,-n[i].s);
            glVertex2d(n[i].s,n[i].s);
            glVertex2d(-n[i].s,n[i].s);
        glEnd();
    }

    glutSwapBuffers();
}

static void key(unsigned char key, int x, int y)
{
    switch (key)
    {
        // close application
        case 27 :   // Esc
        case 'q':
            exit(0);
            break;
        // add random AABB
        case '+':
        case 'a':
            AddAABB();
            break;
        // remove last AABB
        case '-':
        case 'r':
            RemoveAABB();
            break;
        // pause
        case 13 :   // Enter
        case ' ':
            pause = !pause;
            break;
    }

    glutPostRedisplay();
}

static void idle(void)
{
    glutPostRedisplay();
}

int main(int argc, char *argv[])
{
    glutInit(&argc, argv);
    glutInitWindowSize(600,480);
    glutInitWindowPosition(10,10);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);

    glutCreateWindow("c05 - Uniform Grid and Pair Management");

    glutReshapeFunc(resize);
    glutDisplayFunc(display);
    glutKeyboardFunc(key);
    glutIdleFunc(idle);

    glClearColor(1,1,1,1);

    srand(time(NULL));
    init();
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    glutMainLoop();

    return EXIT_SUCCESS;
}
